//
//  VILoaderImageView.m
//  LoaderImageView
//
//  Created by Anthony Alesia on 5/17/12.
//  Copyright (c) 2012 VOKAL. All rights reserved.
//

#import "VILoaderImageView.h"
#import "UIImage+Decompression.h"

#define TMP NSTemporaryDirectory()

//Set to -1 to never recache a cached image.

#define LENGTH_OF_CACHE                 86400

#define PREDECOMPRESS                   1

#define LOCAL_CACHE_MAX_ITEMS           50

#define MAX_CONCURRENT_OPERATIONS       5

@interface VILoaderImageView ()

+ (NSOperationQueue *)getQueue;
@property UIActivityIndicatorView *activityIndicator;
@property UIImage *defaultImage;
@property NSUInteger uniqueHash;

+ (UIImage *)getCachedImage:(NSString *)imageURLString;
+ (void)cacheImage:(NSString *)imageURLString completion:(void (^)(UIImage *image))completion;
@end

@implementation VILoaderImageView

@synthesize activityIndicator = _activityIndicator;

static NSOperationQueue *_queue = nil;

+ (NSOperationQueue *)getQueue
{
    static dispatch_once_t onceToken;
    dispatch_once(&onceToken, ^{
        _queue = [[NSOperationQueue alloc] init];
        [_queue setMaxConcurrentOperationCount:MAX_CONCURRENT_OPERATIONS];
    });

    return _queue;
}

#pragma mark - Public Methods

- (id)initWithFrame:(CGRect)frame
{
    self = [super initWithFrame:frame];
    if (self) {
        self.backgroundColor = [UIColor clearColor];
        self.clipsToBounds = YES;
        self.opaque = NO;
    }
    return self;
}

#pragma mark - Instance Methods

- (BOOL)isCancelled:(NSString *)urlString
{
    return self.uniqueHash != [urlString hash];
}

- (void)setImageUrl:(NSString *)imageUrl defaultImage:(UIImage *)defaultImage animated:(BOOL)animated
{
    self.defaultImage = self.image = defaultImage;
    [self setImageUrl:imageUrl animated:animated];
}

- (void)setImageUrl:(NSString *)imageUrl animated:(BOOL)animated
{
    self.uniqueHash = [imageUrl hash];
    if (!imageUrl) {
        return;
    }
    dispatch_sync(dispatch_get_global_queue(DISPATCH_QUEUE_PRIORITY_BACKGROUND, 0), ^{
        UIImage *image = [VILoaderImageView getCachedImage:imageUrl];
        if (image != nil) {
            [self setImageOnMainThread:image animated:animated];
        } else {
            [self fetchRemoteImage:imageUrl animated:animated];
        }
    });
}

- (void)setImageUrlClearingCache:(NSString *)imageUrl animated:(BOOL)animated
{
    self.uniqueHash = [imageUrl hash];
    [[VILoaderImageView getQueue] addOperationWithBlock:^{
        if (_activityIndicator == nil) {
            [self startIndicatorAnimating];
        }

        [VILoaderImageView recacheImage:imageUrl completion:^(UIImage *image) {
            [self stopIndicatorAnimating];
            if (![self isCancelled:imageUrl]) {
                return;
            }

            if (animated) {
                [self animateImage:image];
            } else {
                self.image = image;
            }
        }];
    }];
}

- (void)fetchRemoteImage:(NSString *)imageUrl animated:(BOOL)animated
{
    [[VILoaderImageView getQueue] addOperationWithBlock:^{
        if (_activityIndicator == nil) {
            [self startIndicatorAnimating];
        }

        [VILoaderImageView cacheImage:imageUrl completion:^(UIImage *image) {
            [self stopIndicatorAnimating];
            if (![self isCancelled:imageUrl]) {
                [self setImageOnMainThread:image animated:animated];
            }
        }];
    }];

}

- (void)setImageOnMainThread:(UIImage *)image animated:(BOOL)animated
{
    dispatch_async(dispatch_get_main_queue(), ^{
        if (nil == image)
            self.image = self.defaultImage;
        else if (animated) {
            [self animateImage:image];
        } else {
            self.image = image;
        }
    });
}

- (void)startIndicatorAnimating
{
    NSLog(@"startIndicatorAnimating called");
    dispatch_async(dispatch_get_main_queue(), ^{
        if (nil == _activityIndicator) {
            _activityIndicator = [[UIActivityIndicatorView alloc] initWithFrame:CGRectMake(0, 0,
                                                                                           self.frame.size.width,
                                                                                           self.frame.size.height)];
            _activityIndicator.activityIndicatorViewStyle = UIActivityIndicatorViewStyleWhite;
        }
        [_activityIndicator startAnimating];
        [self addSubview:_activityIndicator];
    });
}

- (void)stopIndicatorAnimating
{
    NSLog(@"stoptIndicatorAnimating called");
    dispatch_async(dispatch_get_main_queue(), ^{
        [self.activityIndicator stopAnimating];
        [self.activityIndicator removeFromSuperview];
        self.activityIndicator = nil;
    });
}

#pragma mark - Private Methods

#pragma mark - Instance Methods

- (void)animateImage:(UIImage *)image
{
    dispatch_async(dispatch_get_main_queue(), ^{
        self.image = image;
        CATransition *animation = [CATransition animation];
        [animation setDuration:0.2];
        [animation setType:kCATransitionFade];
        [animation setSubtype:kCATransitionFade];

        [[self layer] addAnimation:animation forKey:@"DisplayView"];

        self.image = image;

        self.frame = CGRectMake(self.frame.origin.x, self.frame.origin.y, self.frame.size.width+1, self.frame.size.height+1);
        self.frame = CGRectMake(self.frame.origin.x, self.frame.origin.y, self.frame.size.width-1, self.frame.size.height-1);
    });
}

#pragma mark - Class Methods
+ (void)recacheImage:(NSString*)imageURLString completion:(void (^)(UIImage *image))completion
{
    [self cacheImage:imageURLString deleteExisting:YES completion:completion];
}

+ (void)cacheImage:(NSString *)imageURLString completion:(void (^)(UIImage *image))completion
{
    [self cacheImage:imageURLString deleteExisting:NO completion:completion];
}

+ (void)cacheImage:(NSString *)imageURLString deleteExisting:(BOOL)deleteExisting completion:(void (^)(UIImage *))completion
{
    dispatch_async(dispatch_get_global_queue(DISPATCH_QUEUE_PRIORITY_BACKGROUND, 0), ^{
        NSURL *imageURL = [NSURL URLWithString:imageURLString];

        // Generate a unique path to a resource representing the image you want
        NSString *uniquePath = [self getUniquePathFromURL:imageURLString];

        UIImage *image;

        NSFileManager *fm = [NSFileManager defaultManager];

        // Check for file existence
        if(![fm fileExistsAtPath:uniquePath]) {
            image = [VILoaderImageView writeFile:imageURL toPath:uniquePath];

        } else {

            NSDictionary *attrs = [fm attributesOfItemAtPath:uniquePath error:nil];
            NSDate *date = (NSDate*)[attrs objectForKey: NSFileCreationDate];

            BOOL deleteBecauseTooOld = [date timeIntervalSinceNow] < -LENGTH_OF_CACHE && LENGTH_OF_CACHE != -1;
            
            if (deleteBecauseTooOld || deleteExisting){
                [fm removeItemAtPath:uniquePath error:nil];
                image = [VILoaderImageView writeFile:imageURL toPath:uniquePath];
            } else {
                image = [VILoaderImageView getCachedImage:uniquePath];
            }
        }

        dispatch_async(dispatch_get_main_queue(), ^{
            if (completion) {
                completion(image);
            }
        });
    });
}

+ (UIImage*)writeFile:(NSURL*)imageURL toPath:(NSString*)uniquePath
{
    NSData *data = [NSData dataWithContentsOfURL:imageURL];
    if (nil == data)
        return nil;

    UIImage* image = [UIImage imageWithData:data];
    //Rescale as a retina image.
    image = [UIImage imageWithCGImage:[image CGImage] scale:2.0 orientation:UIImageOrientationUp];

    if([[imageURL absoluteString] rangeOfString:@".jpg" options:NSCaseInsensitiveSearch].location != NSNotFound || [[imageURL absoluteString] rangeOfString:@".jpeg" options:NSCaseInsensitiveSearch].location != NSNotFound) {
        [UIImageJPEGRepresentation(image, .8) writeToFile:uniquePath atomically: YES];
    } else {
        [UIImagePNGRepresentation(image) writeToFile:uniquePath atomically: YES];

    }

    return image;
}

+ (UIImage *)getCachedImage:(NSString *)imageURLString
{
    NSString *uniquePath = [self getUniquePathFromURL:imageURLString];
    UIImage *image = nil;

    if([[NSFileManager defaultManager] fileExistsAtPath: uniquePath]) {

        image = (PREDECOMPRESS ? [[UIImage alloc] initWithContentsOfFileDecompressed: uniquePath] : [UIImage imageWithContentsOfFile: uniquePath]);
        image = [UIImage imageWithCGImage:[image CGImage] scale:4.0 orientation:UIImageOrientationUp];
    }

    return image;
}

+ (void)getCachedImage:(NSString *)imageURLString completion:(void (^)(UIImage *image))completion
{
    UIImage *image = [VILoaderImageView getCachedImage:imageURLString];

    if (image == nil) {
        [VILoaderImageView cacheImage:imageURLString completion:^(UIImage *image) {
            [[NSOperationQueue mainQueue] addOperationWithBlock:^{
                completion(image);
            }];
        }];
    } else {
        completion(image);
    }
}

+ (NSString *)getUniquePathFromURL:(NSString*)imageURLString
{
    NSString *filename = [[imageURLString stringByReplacingOccurrencesOfString:@":" withString:@""]
                          stringByReplacingOccurrencesOfString:@"/" withString:@""];

    NSString* uniquePath =  [TMP stringByAppendingPathComponent: filename];
    uniquePath = [uniquePath stringByAppendingString:@"@2x"];
    
    return uniquePath;
}

@end
